@@ -466,11 +466,7 @@ static void __alx_set_rx_mode(struct net_device *netdev)
 
 	if (!(netdev->flags & IFF_ALLMULTI)) {
 		netdev_for_each_mc_addr(ha, netdev)
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
-			alx_add_mc_addr(hw, ha->addr, mc_hash);
-#else
-			alx_add_mc_addr(hw, ha->dmi_addr, mc_hash);
-#endif
+			alx_add_mc_addr(hw, mc_addr(ha), mc_hash);
 
 		alx_write_mem32(hw, ALX_HASH_TBL0, mc_hash[0]);
 		alx_write_mem32(hw, ALX_HASH_TBL1, mc_hash[1]);
@@ -1391,7 +1387,6 @@ static void alx_remove(struct pci_dev *pdev)
 
 	pci_disable_pcie_error_reporting(pdev);
 	pci_disable_device(pdev);
-	pci_set_drvdata(pdev, NULL);
 
 	free_netdev(alx->dev);
 }
